"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class


class WASystem(ABC):
    """Used to manage simulation parameters and can be used to update simulation dynamics

    A system is used for organizational purposes almost exclusively. It is passed to other
    module classes to handle passing various attributes. Could be responsible for updating
    module dynamics depending on the underlying models.

    Args:
        step_size (double): the step size for the simulation
        render_step_size (double, optional): Render step size. Defaults to 2e-2.

    Attributes:
        step_number (int): Counter of the Advance function
        step_size (double): the step size for the simulation
        render_step_size (double): the render step size
        time (double): stores the time from the simulation
    """

    def __init__(self, step_size, render_step_size=2e-2):
        self.step_number = 0
        self.step_size = step_size
        self.render_step_size = render_step_size

        self.time = 0

    def advance(self):
        """Advance the system

        Will update the time and increment the step number
        """
        self.time += self.step_size
        self.step_number += 1

    def get_sim_time(self):
        """Get the simulation time

        Returns:
            double: the simulation time
        """
        return self.time

    def get_step_number(self):
        """Get the step number

        Returns:
            int: the step number
        """
        return self.step_number



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.terrain import WATerrain

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh


def read_terrain_model_file(filename):
    """Get the json specification file describing a ChTerrain

    Args:
        filename (str): the relative path to a terrain json file

    Returns:
        str: the absolute json file
    """
    import json

    full_filename = chrono.GetChronoDataFile(filename)

    with open(full_filename) as f:
        j = json.load(f)

    if "Terrain" not in j:
        print("'Terrain' not present in the passed json file.")
        exit()
    terrain_filename = chrono.GetChronoDataFile(j["Terrain"]["Input File"])

    return terrain_filename


class WAChronoTerrain(WATerrain):
    """Wrapper for a ChTerrain

    Only support for RigidTerrain is supported for now.

    Args:
        filename (str): the json specification file describing the terrain
        system (WASystem): system used to create the terrain

    Attributes:
        terrain (RigidTerrain): the created rigid terrain
    """

    def __init__(self, filename, system):
        # Get the filenames
        terrain_filename = read_terrain_model_file(filename)

        # Create the terrain
        self.terrain = veh.RigidTerrain(system.system, terrain_filename)

    def synchronize(self, time):
        """Synchronize the terrain at the specified time

        Args:
            time (double): time at which synchronization should occur
        """
        self.terrain.Synchronize(time)

    def advance(self, step):
        """Advance the terrain dynamics by the specified step

        Args:
            step (double): the step at which the simulation should be updated by
        """
        self.terrain.Advance(step)


"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.visualization import WAMatplotlibVisualization
from wa_simulator.inputs import WAVehicleInputs
from wa_simulator.vector import WAVector

# Other imports
import sys
import numpy as np


class WAController(ABC):
    """Base class for a controller

    Controllers are responsible for outputing a steering, throttle and braking value.
    This is done because in real life, those are the inputs our cars will have. The
    derived controller's (i.e. the new class that inherits from this class)
    responsibility is to take inputs from the simulation and return these values
    through the get_inputs method.

    Args:
        system (ChSystem): The system used to manage the simulation

    Attributes:
        inputs (WAVehicleInputs): Inputs to the vehicle model
    """

    def __init__(self, system):
        self.system = system
        self.inputs = WAVehicleInputs()

    @abstractmethod
    def synchronize(self, time):
        """Synchronize the controller at the specified time

        Function is primarily as a semantic separation between different functionality.
        Most of the time, all controller logic can be placed in the Advance method. ROS would
        be a good example of an element that would publish in the Synchronize method and have
        other logic in the Advance method.

        Args:
                time (double): the time at which the controller should synchronize all depends to
        """
        pass

    @abstractmethod
    def advance(self, step):
        """Advance the controller by the specified step

        Args:
                step (double): the time step at which the controller should be advanced
        """
        pass

    def get_inputs(self):
        """Get the vehicle inputs

        Returns:
                WAVehicleInputs: The input class
        """
        return self.inputs

    def _get_steering(self):
        return self.inputs.steering

    def _get_throttle(self):
        return self.inputs.throttle

    def _get_braking(self):
        return self.inputs.braking

    def _set_steering(self, steering):
        self.inputs.steering = steering

    def _set_throttle(self, throttle):
        self.inputs.throttle = throttle

    def _set_braking(self, braking):
        self.inputs.braking = braking

    steering = property(_get_steering, _set_steering)
    throttle = property(_get_throttle, _set_throttle)
    braking = property(_get_braking, _set_braking)


class WASimpleController(WAController):
    """Simple controller designed to never change the inputs

    Can be used for situations where controlling the vehicle isn't actually necessary
    """

    def advance(self, step):
        pass

    def synchronize(self, time):
        pass


class _WAKeyboardController(WAController):
    """Base keyboard controller. Still must be inherited (can't be instantiated). Has utilites.

    Args:
        system (ChSystem): The system used to manage the simulation
        steering_target (double): The target steering value.
        throttle_target (double): The target throttle value.
        braking_target (double): The target braking value.
        steering_delta (double): The delta steering value.
        throttle_delta (double): The delta throttle value.
        braking_delta (double): The delta braking value.
        steering_gain (double): The gain steering value.
        throttle_gain (double): The gain throttle value.
        braking_gain (double): The gain braking value.
    """

    def __init__(self, system):
        super().__init__(system)

        self.steering_target = 0
        self.throttle_target = 0
        self.braking_target = 0

        self.steering_delta = system.render_step_size / 2.0
        self.throttle_delta = system.render_step_size / 6.0
        self.braking_delta = system.render_step_size / 1.5

        self.steering_gain = 4.0
        self.throttle_gain = 4.0
        self.braking_gain = 4.0

    def set_steering_delta(self, steering_delta):
        """Sets the steering delta value

        Args:
            steering_delta (double): the new steering delta value
        """
        self.steering_delta = steering_delta

    def set_throttle_delta(self, throttle_delta):
        """Sets the throttle delta value

        Args:
            throttle_delta (double): the new throttle delta value
        """
        self.throttle_delta = throttle_delta

    def set_braking_delta(self, braking_delta):
        """Sets the braking delta value

        Args:
            braking_delta (double): the new braking delta value
        """
        self.braking_delta = braking_delta

    def set_gains(steering_gain, throttle_gain, braking_gain):
        """Sets the controllers gains

        Args:
            steering_gain (double): the new steering gain
            throttle_gain (double): the new throttle gain
            braking_gain (double): the new braking gain
        """
        self.steering_gain = steering_gain
        self.throttle_gain = throttle_gain
        self.braking_gain = braking_gain

    def advance(self, step):
        """Advance the controller by the specified step

        Integrates dynamics over some step range. If the original step is the same as the passed
        step value, the method is only run once.

        Args:
                step (double): the time step at which the controller should be advanced
        """
        # Integrate dynamics, taking as many steps as required to reach the value 'step'
        t = 0
        while t < step:
            h = step - t

            steering_deriv = self.steering_gain * \
                (self.steering_target - self.steering)
            throttle_deriv = self.throttle_gain * \
                (self.throttle_target - self.throttle)
            braking_deriv = self.braking_gain * \
                (self.braking_target - self.braking)

            self.steering += min(self.steering_delta, h *
                                 steering_deriv, key=abs)
            self.throttle += min(self.throttle_delta, h *
                                 throttle_deriv, key=abs)
            self.braking += min(self.braking_delta, h * braking_deriv, key=abs)

            t += h

    def update(self, key):
        """Update the target values based on the key.

        The updated target values are based off the delta values for that respective input.
        In this controller, input values are clipped at [-1,1] or [0,1].

        0: throttle increase, braking decreases
        1: adjust steering right
        2: brake increase, throttle decrease
        3: adjust steering left

        Args:
            key (int): the key input
        """
        if key == -1:
            return
        elif key == 0:
            self.throttle_target = np.clip(
                self.throttle_target + self.throttle_delta, 0.0, +1.0
            )
            if self.throttle_target > 0:
                self.braking_target = np.clip(
                    self.braking_target - self.braking_delta * 3, 0.0, +1.0
                )
        elif key == 2:
            self.throttle_target = np.clip(
                self.throttle_target - self.throttle_delta * 3, 0.0, +1.0
            )
            if self.throttle_target <= 0:
                self.braking_target = np.clip(
                    self.braking_target + self.braking_delta, 0.0, +1.0
                )
        elif key == 1:
            self.steering_target = np.clip(
                self.steering_target + self.steering_delta, -1.0, +1.0
            )
        elif key == 3:
            self.steering_target = np.clip(
                self.steering_target - self.steering_delta, -1.0, +1.0
            )
        else:
            print("Key not recognized")
            return


try:
    import termios
    import tty
    import atexit
    from select import select

    class WATerminalKeyboardController(_WAKeyboardController):
        """Controls a vehicle via input from the terminal window.

        Uses the KeyGetter object to grab input from the user in the terminal window.
        Inherits from the _WAKeyboardController method

        Args:
            system (ChSystem): The system used to manage the simulation

        Attributes:
            key_getter (KeyGetter): The object used to grab input from the terminal
        """

        class KeyGetter:
            """Gets user input from the terminal.

            Will look for different input from the command line. The terminal window
            must be active for this to work.

            TODO: Fairly confident this only works on UNIX.

            Attributes:
                    fd (STDIN_FILENO): Integer file descriptor.
                    new_term (list): tty attributes for the fd. Setting terminal settings.
                    old_term (list): tty attributes for the fd. Restoring terminal settings.
            """

            def __init__(self):
                # Save the terminal settings
                self.fd = sys.stdin.fileno()
                self.new_term = termios.tcgetattr(self.fd)
                self.old_term = termios.tcgetattr(self.fd)

                # New terminal setting unbuffered
                self.new_term[3] = self.new_term[3] & ~termios.ICANON & ~termios.ECHO
                termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.new_term)

                # Support normal-terminal reset at exit
                atexit.register(self.set_normal_term)

            def __call__(self):
                """Checks the terminal window for a user input.

                This method is called through key_getter_object(). Will check terminal
                window for arrow keys. Will exit if any other key is pressed.

                Returns:
                    int: a value between [0,3] describing the arrow key pressed
                """
                dr, dw, de = select([sys.stdin], [], [], 0)
                if dr == []:
                    return -1

                c = sys.stdin.read(3)[2]
                vals = [65, 67, 66, 68]

                if vals.count(ord(c)) == 0:
                    self.set_normal_term()
                    raise RuntimeError(f'"{ord(c)}" is not an arrow key.')

                return vals.index(ord(c))

            def set_normal_term(self):
                """Resets to normal terminal.  On Windows this is a no-op."""
                termios.tcsetattr(self.fd, termios.TCSAFLUSH, self.old_term)

        def __init__(self, system):
            super().__init__(system)

            self.key_getter = self.KeyGetter()

        def key_check(self):
            """Get the key from the KeyGetter and update target values based on input."""
            try:
                key = self.key_getter()
                self.update(key)
            except Exception as e:
                print(e)
                return

        def synchronize(self, time):
            """Synchronize the controller at the specified time

            Calls KeyCheck

            Args:
                    time (double): the time at which the controller should synchronize all depends to
            """
            self.key_check()
except:
    pass


class WAMultipleControllers(WAController):
    """Wrapper class for multiple controllers. Allows multiple controllers to be used.

    The input values for the model are grabbed from the first controller in the list.

    Args:
        controllers (list): List of controllers.
    """

    def __init__(self, controllers):
        self.controllers = controllers

    def synchronize(self, time):
        """Synchronize each controller at the specified time

        Args:
            time (double): the time at which the controller should synchronize all modules
        """
        for ctr in self.controllers:
            ctr.synchronize(time)

    def advance(self, step):
        """Advance the state of each managed controller

        Args:
            step (double): the time step at which the controller should be advanced
        """
        for ctr in self.controllers:
            ctr.Advance(step)

    def get_inputs(self):
        """Get the vehicle inputs

        Overrides base class method. Will just grab the first controllers inputs.

        Returns:
                The input class
        """

        return self.controllers[0].get_inputs()


class WAPIDController(WAController):
    """PID Controller that contains a lateral and longitudinal controller

    Uses the lateral controller for steering and longitudinal controller throttle/braking

    Args:
        system (ChSystem): The system used to manage the simulation
        vehicle (WAVehicle): The vehicle to grab the state from
        path (WAPath): The path to follow
        lat_controller (WAPIDLateralController, optional): Lateral controller for steering. Defaults to None.
        long_controller (WAPIDLongitudinalController, optional): Longitudinal controller for throttle/braking. Defaults to None.

    Attributes:
        vehicle (WAVehicle): The vehicle to grab the state from
        path (WAPath): The path to follow
        lat_controller (WAPIDLateralController): Lateral controller for steering
        long_controller (WAPIDLongitudinalController): Longitudinal controller for throttle/braking
    """

    def __init__(self, system, vehicle, path, lat_controller=None, long_controller=None):
        super().__init__(system)

        self.vehicle = vehicle
        self.path = path

        # Lateral controller (steering)
        if lat_controller is None:
            lat_controller = WAPIDLateralController(system, vehicle, path)
            lat_controller.set_gains(Kp=0.4, Ki=0, Kd=0)
            lat_controller.set_lookahead_distance(dist=5)
        self.lat_controller = lat_controller

        if long_controller is None:
            # Longitudinal controller (throttle and braking)
            long_controller = WAPIDLongitudinalController(system, vehicle)
            long_controller.set_gains(Kp=0.4, Ki=0, Kd=0)
            long_controller.set_target_speed(speed=10.0)
        self.long_controller = long_controller

        self.target_steering = 0
        self.target_throttle = 0
        self.target_braking = 0

        self.steering_delta = 1.0 / 50
        self.throttle_delta = 1.0 / 50
        self.braking_delta = 1.0 / 50

        self.steering_gain = 4.0
        self.throttle_gain = 0.25
        self.braking_gain = 4.0

    def set_delta(self, steering_delta, throttle_delta, braking_delta):
        """Set the delta values

        Args:
            steering_delta (float): max steering delta
            throttle_delta (float): max throttle delta
            braking_delta (float): max braking delta
        """
        self.steering_delta = steering_delta
        self.throttle_delta = throttle_delta
        self.braking_delta = braking_delta

    def set_gains(self, steering_gain, throttle_gain, braking_gain):
        """Set the gain values

        Args:
            steering_gain (float): steering gain
            throttle_gain (float): throttle gain
            braking_gain (float): braking gain
        """
        self.steering_gain = steering_gain
        self.throttle_gain = throttle_gain
        self.braking_gain = braking_gain

    def synchronize(self, time):
        """Synchronize each controller at the specified time

        Args:
            time (double): the time at which the controller should synchronize all modules
        """
        self.lat_controller.synchronize(time)
        self.long_controller.synchronize(time)

    def advance(self, step):
        """Advance the state of each controller

        Args:
            step (double): the time step at which the controller should be advanced
        """
        self.lat_controller.advance(step)
        self.long_controller.advance(step)

        self.target_steering = self.lat_controller.steering
        self.target_throttle = self.long_controller.throttle
        self.target_braking = self.long_controller.braking

        # Integrate dynamics, taking as many steps as required to reach the value 'step'
        t = 0.0
        while t < step:
            h = min(self.system.step_size, step - t)

            steering_deriv = self.steering_gain * \
                (self.target_steering - self.steering)
            throttle_deriv = self.throttle_gain * \
                (self.target_throttle - self.throttle)
            braking_deriv = self.braking_gain * \
                (self.target_braking - self.braking)

            self.steering += min(h * steering_deriv,
                                 self.steering_delta, key=abs)
            self.throttle += min(h * throttle_deriv,
                                 self.throttle_delta, key=abs)
            self.braking += min(h * braking_deriv, self.braking_delta, key=abs)

            t += h

    def get_inputs(self):
        """Get the vehicle inputs

        Overrides base class method. Grabs the steering from the lateral controller and the
        throttle and braking from the longitudinal controller

        Returns:
            WAVehicleInputs: The input class
        """
        return WAVehicleInputs(
            self.steering,
            self.throttle,
            self.braking,
        )


class WAPIDLateralController(WAController):
    """Lateral (steering) controller which minimizes error using a PID

    Args:
        system (ChSystem): The system used to manage the simulation
        vehicle (WAVehicle): the vehicle who has dynamics
        path (WAPath): the path the vehicle is attempting to follow

    Attributes:
        vehicle (WAVehicle): the vehicle who has dynamics
        path (WAPath): the path the vehicle is attempting to follow
        Kp (double): proportional gain
        Ki (double): integral gain
        Kd (double): derivative gain
        dist (double): lookahead distance
        target (double): point on the path that is attempting to reach
        sentinel (double): some point a dist directly in front of the vehicle
        err (double): overall error
        errd (double): derivative error (not accumulated)
        erri (double): integral error accumulated over time
    """

    def __init__(self, system, vehicle, path):
        super().__init__(system)

        self.Kp = 0
        self.Ki = 0
        self.Kd = 0

        self.dist = 0
        self.target = WAVector([0, 0, 0])
        self.sentinel = WAVector([0, 0, 0])

        self.err = 0
        self.errd = 0
        self.erri = 0

        self.path = path
        self.vehicle = vehicle

    def set_gains(self, Kp, Ki, Kd):
        """Set the gains

        Args:
            Kp (double): new proportional gain
            Ki (double): new integral gain
            Kd (double): new derivative gain
        """
        self.Kp = Kp
        self.Ki = Ki
        self.Kd = Kd

    def set_lookahead_distance(self, dist):
        """Set the lookahead distance

        Args:
            dist (double): new lookahead distance
        """
        self.dist = dist

    def synchronize(self, time):
        """Synchronize the controller at the passed time

        Doesn't actually do anything.

        Args:
            time (double): the time to synchronize the controller to
        """
        pass

    def advance(self, step):
        """Advance the state of the controller by step

        Args:
            step (double): step size to update the controller by
        """
        x, y, yaw, v = self.vehicle.get_simple_state()
        self.sentinel = WAVector(
            [
                self.dist * np.cos(yaw) + x,
                self.dist * np.sin(yaw) + y,
                0,
            ]
        )

        self.target = self.path.calc_closest_point(self.sentinel)

        # The "error" vector is the projection onto the horizontal plane (z=0) of
        # vector between sentinel and target
        err_vec = self.target - self.sentinel
        err_vec.z = 0

        # Calculate the sign of the angle between the projections of the sentinel
        # vector and the target vector (with origin at vehicle location).
        sign = self.calc_sign(x, y)

        # Calculate current error (magnitude)
        err = sign * err_vec.length()

        # Estimate error derivative (backward FD approximation).
        self.errd = (err - self.err) / step

        # Calculate current error integral (trapezoidal rule).
        self.erri += (err + self.err) * step / 2

        # Cache new error
        self.err = err

        # Return PID output (steering value)
        self.steering = np.clip(
            self.Kp * self.err + self.Ki * self.erri + self.Kd * self.errd, -1.0, 1.0
        )

    def calc_sign(self, x, y):
        """Calculate the sign of the angle between the projections of the sentinel vector
        and the target vector (with origin at vehicle location).

        Args:
            x (float): x position
            y (float): y position

        Returns:
            int: the sign indicating direction of the state and the sentinel on the path (-1 for left or 1 for right)
        """
        pos = WAVector([x, y, 0])

        sentinel_vec = self.sentinel - pos
        target_vec = self.target - pos

        temp = np.dot(np.cross(sentinel_vec, target_vec), WAVector([0, 0, 1]))

        return int(temp > 0) - int(temp < 0)


class WAPIDLongitudinalController(WAController):
    """Longitudinal (throttle, braking) controller which minimizes error using a PID

    Args:
        system (ChSystem): The system used to manage the simulation
        vehicle (WAVehicle): the vehicle who has dynamics

    Attributes:
        vehicle (WAVehicle): the vehicle who has dynamics
        Kp (double): proportional gain
        Ki (double): integral gain
        Kd (double): derivative gain
        err (double): overall error
        errd (double): derivative error (not accumulated)
        erri (double): integral error accumulated over time
        speed (double): the current speed
        target_speed (double): the target speed
        throttle_threshold (double): throttle position at which vehicle is moving too fast and the step size should decrease
    """

    def __init__(self, system, vehicle):
        super().__init__(system)

        self.Kp = 0
        self.Ki = 0
        self.Kd = 0

        self.err = 0
        self.errd = 0
        self.erri = 0

        self.speed = 0
        self.target_speed = 0

        self.throttle_threshold = 0.2

        self.vehicle = vehicle

    def set_gains(self, Kp, Ki, Kd):
        """Set the gains

        Args:
            Kp (double): new proportional gain
            Ki (double): new integral gain
            Kd (double): new derivative gain
        """
        self.Kp = Kp
        self.Ki = Ki
        self.Kd = Kd

    def set_target_speed(self, speed):
        """Set the target speed for the controller

        Args:
            speed (double): the new target speed
        """
        self.target_speed = speed

    def synchronize(self, time):
        """Synchronize the controller at the passed time

        Doesn't actually do anything.

        Args:
            time (double): the time to synchronize the controller to
        """
        pass

    def advance(self, step):
        """Advance the state of the controller by step

        Args:
            step (double): step size to update the controller by
        """

        _, _, _, self.speed = self.vehicle.get_simple_state()

        # Calculate current error
        err = self.target_speed - self.speed

        # Estimate error derivative (backward FD approximation)
        self.errd = (err - self.err) / step

        # Calculate current error integral (trapezoidal rule).
        self.erri += (err + self.err) * step / 2

        # Cache new error
        self.err = err

        # Return PID output (steering value)
        throttle = np.clip(
            self.Kp * self.err + self.Ki * self.erri + self.Kd * self.errd, -1.0, 1.0
        )

        if throttle > 0:
            # Vehicle moving too slow
            self.braking = 0
            self.throttle = throttle
        elif self.throttle > self.throttle_threshold:
            # Vehicle moving too fast: reduce throttle
            self.braking = 0
            self.throttle += throttle
        else:
            # Vehicle moving too fast: apply brakes
            self.braking = -throttle
            self.throttle = 0


class WAMatplotlibController(_WAKeyboardController):
    """Controls a vehicle via keyboard input from a matplotlib figure

    Will asynchronously change inputs based on user input to the matplotlib window.

    Args:
        system (ChSystem): The system used to manage the simulation
        vis (WAMatplotlibVisualization): The visualization that holds a matplotlib figure

Attributes:
        vis (WAMatplotlibVisualization): The visualization that holds a matplotlib figure
    """

    def __init__(self, system, vis):
        if not isinstance(vis, WAMatplotlibVisualization):
            raise TypeError(
                "Visualization passed in is not a WAMatplotlibVisualization."
            )

        super().__init__(system)

        vis.register_key_press_event(self.key_press)

        self._input_dict = {'up': 0, 'right': 1, 'down': 2, 'left': 3}

    def synchronize(self, time):
        """Synchronize the controller at the specified time

        Doesn't do anything since this controller is completely asynchronous

        Args:
                time (double): the time at which the controller should synchronize all depends to
        """
        pass

    def key_press(self, value):
        self.update(self._input_dict[value])



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import argparse


class WAArgumentParser(argparse.ArgumentParser):
    def __init__(
        self, use_sim_defaults=True, description=("Wisconsin Autonomous Simulator")
    ):
        """Argument parser wrapper.

        Has handy default arguments method

        Args:
            use_sim_defaults (bool, optional): Use the default arguments (can add more). Defaults to True.
            description (str, optional): Description used to describe the simulation. Printed by the help menu. Defaults to ("Wisconsin Autonomous Simulator").
        """
        super().__init__(description=description)

        if use_sim_defaults:
            self.add_default_sim_arguments()

    def add_default_sim_arguments(self):
        """Adds default arguments for most simulations

        Includes:
            -s,--sim_step_size: Simulation step size
            -rs,--render_step_size: Rendering step size
            -iv,--irrlicht: Use irrlicht visualization
            -mv,--matplotlib: Use matplotlib visualization
            -r,--record: Record the simulation to a csv
        """
        self.add_argument(
            "-s",
            "--sim_step_size",
            type=float,
            help="Simulation Step Size",
            default=3e-3,
            dest="step_size",
        )

        # Visualization
        self.add_argument(
            "-rs",
            "--render_step_size",
            type=float,
            help="Render Update Rate [Hz]",
            default=1 / 10.0,
        )
        self.add_argument(
            "-iv",
            "--irrlicht",
            action="store_true",
            help="Use Irrlicht to Visualize",
            default=False,
        )
        self.add_argument(
            "-mv",
            "--matplotlib",
            action="store_true",
            help="Use Matplotlib to Visualize",
            default=False,
        )

        self.add_argument(
            "-r",
            "--record",
            action="store_true",
            help="Record Simple State Data",
            default=False,
        )



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.data_loader import get_wa_data_file
from wa_simulator.constants import GRAVITY

# Other imports
import numpy as np


def load_properties_from_json(filename, prop):
    """Load a specified property from a json specification file

    Will load a json file and extract the passed property field for use
    by the underyling vehicle object

    Args:
        filename (str): the filename location within the set WA data folder for the json file
        property (str): the property to get. Ex: "Vehicle Properties"

    Raises:
        ValueError: The property field isn't found

    Returns:
        dict: the property field extracted from the json file
    """
    import json

    full_filename = get_wa_data_file(filename)

    with open(full_filename) as f:
        j = json.load(f)

    if prop not in j:
        raise ValueError(f"{prop} not found in json.")

    return j[prop]


class WAVehicle(ABC):
    """Base class for a WAVehicle.

    To implement a new vehicle model, override this class. A WAVehicle should interact
    with the terrain/assets/world and take three inputs: steering, throttle, braking.

    Args:
        filename (str, optional): Filename to be used for visualization properties

    Attributes:
        vis_properties (dict): Visual properties used for visualization of the vehicle.
    """

    def __init__(self, filename=None):
        self.vis_properties = (
            dict()
            if filename is None
            else load_properties_from_json(filename, "Visualization Properties")
        )

    @abstractmethod
    def synchronize(self, time, vehicle_inputs):
        """Synchronize the vehicle at the specified time and driver inputs

        Args:
            time (double): the time at which the synchronize the vehicle to
            vehicle_inputs (WAVehicleInputs): Inputs for the underyling dynamics
        """
        pass

    @abstractmethod
    def advance(self, step):
        """Advance the vehicle by the specified step

        Args:
            step (double): how much to advance the vehicle by
        """
        pass

    @abstractmethod
    def get_simple_state(self):
        """Get a simple state representation of the vehicle.

        Must return a tuple with the following values:
            (x position, y position, yaw about the Z, speed)
        """
        pass


class WALinearKinematicBicycle(WAVehicle):
    """A linear, kinematic bicycle model

    Args:
        filename (str): json specification file used to describe the parameters for the vehicle
        x (double, optional): x position of the vehicle. Defaults to 0.
        y (double, optional): y position of the vehicle. Defaults to 0.
        yaw (double, optional): angle about the Z. Defaults to 0.
        v (double, optional): speed of the vehicle. Defaults to 0.

    Attributes:
        x (double): x position of the vehicle
        y (double): y position of the vehicle
        yaw (double): angle about the Z
        v (double): speed of the vehicle
        omega (double): wheel angular velocity
        omega_dot (double): wheel angular acceleration
        mass (double): vehicle mass
        a (list): 2nd order torque coefficients
        GR (double): gear ratio
        r_eff (double): Effective radius at which the vehicle can turn
        J_e (double): Inertia of the vehicle
        c_a (double): Aerodynamic coefficient
        c_rl (double): Friction coefficient
        c (double): Tire force
        F_max (double): Max force imposed at the wheel
        L (double): Vehicle wheelbase (distance between front and back axles)
        min_steering (double): Minimum steering value allowed
        max_steering (double): Maximum steering value allowed
        min_throttle (double): Minimum throttle value allowed
        max_throttle (double): Maximum throttle value allowed
        min_braking (double): Minimum braking value allowed
        max_braking (double): Maximum braking value allowed
    """

    # Global filenames for vehicle models
    GO_KART_MODEL_FILE = "vehicles/GoKart/GoKart_KinematicBicycle.json"
    IAC_VEH_MODEL_FILE = "vehicles/IAC/IAC_KinematicBicycle.json"

    def __init__(self, filename, x=0, y=0, yaw=0, v=0):
        super().__init__(filename)

        # Simple state variables
        self.x = x
        self.y = y
        self.yaw = yaw
        self.v = v
        self.acc = 0

        # Wheel angular velocity
        self.omega = 0
        self.omega_dot = 0

        self.initialize(load_properties_from_json(
            filename, "Vehicle Properties"))

    def initialize(self, vp):
        # Initialize vehicle properties
        self.mass = vp["Mass"]

        # Torque coefficients
        self.a = vp["Torque Coefficients"]

        # Gear ratio, effective radius and inertia
        self.GR = vp["Gear Ratio"]
        self.r_eff = vp["Effective Radius"]
        self.J_e = vp["Inertia"]

        # Aerodynamic and friction coefficients
        self.c_a = vp["Aerodynamic Coefficient"]
        self.c_rl = vp["Friction Coefficient"]

        # Tire forces
        self.c = vp["C"]
        self.F_max = vp["Max Force"]

        self.L = vp["Wheelbase"]
        (self.min_steering, self.max_steering) = vp["Steering"]
        (self.min_throttle, self.max_throttle) = vp["Throttle"]
        (self.min_braking, self.max_braking) = vp["Braking"]

    def synchronize(self, time, vehicle_inputs):
        """Synchronize the vehicle inputs to the values in this model

        Args:
            time (double): time at which to update the vehicle to
            vehicle_inputs (WAVehicleInputs): vehicle inputs
        """
        s = vehicle_inputs.steering
        t = vehicle_inputs.throttle
        b = vehicle_inputs.braking

        self.steering = np.clip(s, self.min_steering, self.max_steering)
        self.throttle = np.clip(t, self.min_throttle, self.max_throttle)
        self.braking = np.clip(b, self.min_braking, self.max_braking)

        if self.braking > 1e-1:
            self.throttle = -self.braking

    def advance(self, step):
        """Perform a dynamics update

        Args:
            step (double): time step to update the vehicle by
        """
        if self.v == 0 and self.throttle == 0:
            F_x = 0
            F_load = 0
            T_e = 0
        else:
            if self.v == 0:
                # Remove possibity of divide by zero error
                self.v = 1e-8
                self.omega = 1e-8

            # Update engine and tire forces
            F_aero = self.c_a * self.v ** 2
            R_x = self.c_rl * self.v
            F_g = self.mass * GRAVITY * np.sin(0)
            F_load = F_aero + R_x + F_g
            T_e = self.throttle * (
                self.a[0] + self.a[1] * self.omega +
                self.a[2] * self.omega ** 2
            )
            omega_w = self.GR * self.omega  # Wheel angular velocity
            s = (omega_w * self.r_eff - self.v) / self.v  # slip ratio
            cs = self.c * s

            if abs(s) < 1:
                F_x = cs
            else:
                F_x = self.F_max

        if self.throttle < 0:
            if self.v <= 0:
                F_x = 0
            else:
                F_x = -abs(F_x)

        # Update state information
        self.x += self.v * np.cos(self.yaw) * step
        self.y += self.v * np.sin(self.yaw) * step
        self.yaw += self.v / self.L * np.tan(self.steering) * step
        self.v += self.acc * step
        self.acc = (F_x - F_load) / self.mass
        self.omega += self.omega_dot * step
        self.omega_dot = (T_e - self.GR * self.r_eff * F_load) / self.J_e

    def get_simple_state(self):
        """Get a simple state representation of the vehicle.

        Returns:
            tuple: (x position, y position, yaw about the Z, speed)
        """
        return self.x, self.y, self.yaw, self.v



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class


class WAEnvironment(ABC):
    """Base class for the environment object.

    The environment object is responsible for handling data or assets within the world.
    """

    @abstractmethod
    def synchronize(self, time):
        """Synchronize the environment at the specified time

        Function is primarily as a semantic separation between different functionality.
        Most of the time, all environment logic can be placed in the Advance method.

        Args:
                time (double): the time at which the environment should synchronize all depends to
        """
        pass

    @abstractmethod
    def advance(self, step):
        """Advance the environment by the specified step

        Args:
                step (double): the time step at which the environment should be advanced
        """
        pass


class WASimpleEnvironment(WAEnvironment):
    """Simple environment that doesn't have any assets within the world.

    Args:
        filename (str): filename
        sys (WASystem): system
    """

    # Global filenames for environment models
    EGP_ENV_MODEL_FILE = "environments/ev_grand_prix.json"
    IAC_ENV_MODEL_FILE = "environments/iac.json"

    def __init__(self, filename, sys):
        pass

    def synchronize(self, time):
        """Synchronize the environment with the rest of the world at the specified time

        Doesn't actually do anything.

        Args:
            time (double): the time at which the enviornment should be synchronized to
        """
        pass

    def advance(self, step):
        """Advance the state of the environment

        Doesn't actually do anything.

        Args:
            step (double): the time step at which the enviornment should be advanced
        """
        pass



import os
import inspect
import importlib


def _not_hidden(f):
    return f[0] != "_"


def _get_files(file_):
    return [
        os.path.splitext(f.name)[0]
        for f in os.scandir(os.path.dirname(file_))
        if f.is_file() and _not_hidden(f.name)
    ]


def _get_dirs(file_, ignore=[]):
    return [
        os.path.splitext(f.name)[0]
        for f in os.scandir(os.path.dirname(file_))
        if f.is_dir() and _not_hidden(f.name) and f.name not in ignore
    ]


def _import(module, gbls, ignore=[]):
    # Get caller info
    filename = inspect.stack()[1].filename
    path = os.path.dirname(os.path.realpath(filename)).split("wa_simulator")[-1]
    path = f"wa_simulator{path.replace(os.path.sep, '.')}" if path else "wa_simulator"

    # get a handle on the module
    mdl = importlib.import_module(f"{path}.{module}")

    # is there an __all__?  if so respect it
    if "__all__" in mdl.__dict__:
        names = mdl.__dict__["__all__"]
    else:
        # otherwise we import all names that don't begin with _
        names = [x for x in mdl.__dict__ if not x.startswith("_")]

    # now drag them in
    gbls.update({k: getattr(mdl, k) for k in names})



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class


class WATerrain(ABC):
    """Base class for the terrain object

    Maintains knowledge of the terrain or "ground" that a vehicle interacts with.
    """

    @abstractmethod
    def synchronize(self, time):
        """Synchronize the terrain at the specified time

        Function is primarily as a semantic separation between different functionality.
        Most of the time, all terrain logic can be placed in the Advance method.

        Args:
                time (double): the time at which the terrain should synchronize all depends to
        """
        pass

    @abstractmethod
    def advance(self, step):
        """Advance the terrain by the specified step

        Args:
                step (double): the time step at which the terrain should be advanced
        """
        pass



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.vector import WAVector
from wa_simulator.visualization import WAVisualization

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh
import pychrono.irrlicht as irr

# Other imports
import numpy as np
from math import ceil


def draw_path_in_irrlicht(path, sys):
    """Draw a path object in irrlicht

    Args:
        path (WAPath): WA path object to visualize in irrlicht
        sys (WASystem): system that manages the simulation
    """
    ch_points = chrono.vector_ChVectorD()
    for p in np.column_stack((path.x, path.y)):
        ch_points.push_back(chrono.ChVectorD(p[0], p[1], 0.5))
    curve = chrono.ChBezierCurve(ch_points)

    road = sys.system.NewBody()
    road.SetBodyFixed(True)
    sys.system.AddBody(road)

    num_points = len(path.x)
    path_asset = chrono.ChLineShape()
    path_asset.SetLineGeometry(chrono.ChLineBezier(curve))
    path_asset.SetColor(chrono.ChColor(0, 0.8, 0))
    path_asset.SetNumRenderPoints(max(2 * num_points, 400))
    road.AddAsset(path_asset)


def create_sphere_in_irrlicht(system, rgb=(1, 0, 0)):
    """Create and add a sphere to the irrlicht visualization

    Args:
        system (WASystem): the system that manages the simulation
        rgb (tuple, optional): (red, green, blue). Defaults to (1, 0, 0) or red.

    Returns:
        chrono.ChBodyEasySphere: the sphere that was created
    """
    sphere = chrono.ChBodyEasySphere(0.25, 1000, True, False)
    sphere.SetBodyFixed(True)
    sphere.AddAsset(chrono.ChColorAsset(*rgb))
    system.system.Add(sphere)

    return sphere


def update_position_of_sphere(sphere, pos):
    """Update the position of a sphere being visualized in irrlicht

    Args:
        sphere (chrono.ChBodyEasySphere): the sphere to change the position of
        pos (WAVector): the new position
    """
    if not isinstance(pos, WAVector):
        pos = WAVector(pos)

    sphere.SetPos(chrono.ChVectorD(pos.x, pos.y, 0))


class WAChronoIrrlicht(WAVisualization):
    """Chrono irrlicht visualization wrapper. Sets default values for a Vehicle irr app

    Args:
        vehicle (WAChronoVehicle): vehicle that holds the Chrono vehicle
        system (WAChronoSystem): holds information regarding the simulation and performs dynamic updates

    Attributes:
        render_steps (int): steps between which the visualization should update
        system (WAChronoSystem): system used to grab certain parameters of the simulation
        app (ChVehicleIrrApp): The vehicle irrlicht app that displays the 3D rendering
    """

    def __init__(self, vehicle, system):
        self.render_steps = int(
            ceil(system.render_step_size / system.step_size))

        self.system = system

        self.app = veh.ChVehicleIrrApp(vehicle.vehicle)
        self.app.SetHUDLocation(500, 20)
        self.app.SetSkyBox()
        self.app.AddTypicalLogo()
        self.app.AddTypicalLights(
            irr.vector3df(-150.0, -150.0, 200.0),
            irr.vector3df(-150.0, 150.0, 200.0),
            100,
            100,
        )
        self.app.AddTypicalLights(
            irr.vector3df(150.0, -150.0, 200.0),
            irr.vector3df(150.0, 150.0, 200.0),
            100,
            100,
        )
        self.app.SetChaseCamera(chrono.ChVectorD(0.0, 0.0, 1.75), 6.0, 0.5)
        self.app.SetTimestep(system.step_size)

        self.app.AssetBindAll()
        self.app.AssetUpdateAll()

    def advance(self, step):
        """Advance the state of the visualization by the specified step

        Will update the render only if the simulation step is a multiple of render steps

        Args:
            step (double): step size to update the visualization by
        """
        if self.system.get_step_number() % self.render_steps == 0:
            self.app.BeginScene(True, True, irr.SColor(255, 140, 161, 192))
            self.app.DrawAll()
            self.app.EndScene()

        self.app.Advance(step)

    def synchronize(self, time, vehicle_inputs):
        """Synchronize the irrlicht app with the vehicle inputs at the passed time

        Args:
            time (double): time to synchronize the simulation to
            vehicle_inputs (WAVehicleInputs): the vehicle inputs
        """
        d = veh.Inputs()
        d.m_steering = vehicle_inputs.steering
        d.m_throttle = vehicle_inputs.throttle
        d.m_braking = vehicle_inputs.braking
        vehicle_inputs = d

        self.app.Synchronize("", vehicle_inputs)

    def is_ok(self):
        """Checks if the irrlicht rending window is still alive

        Returns:
            bool: whether the simulation is still alive
        """
        return self.app.GetDevice().run()



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.inputs import WAVehicleInputs

# Other imports
from multiprocessing import Process, Pipe, Barrier, Queue, set_start_method
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from IPython.display import display, clear_output


class WAVisualization(ABC):
    """Base class to be used for visualization of the simulation world.

    Derived classes will use various world attributes to visualize the simulation
    """

    @abstractmethod
    def synchronize(self, time, vehicle_inputs):
        """Synchronize the visualization at the specified time with the passed vehicle inputs

        Args:
            time (double): time to synchronize the visualization to
            vehicle_inputs (WAVehicleInputs): inputs to the vehicle. Can be helpful for visualization (debug) purposes.
        """
        pass

    @abstractmethod
    def advance(self, step):
        """Advance the state of the visualization by the specified step

        Args:
            step (double): step size to update the visualization by
        """
        pass

    @abstractmethod
    def is_ok(self):
        """Verifies the visualization is running properly.

        Returns:
            bool: Whether the visualization is running correctly.
        """
        pass


class WAMultipleVisualizations(WAVisualization):
    """Wrapper class for multiple visualizations. Allows multiple visualizations to be used.

    Args:
        visualizations (list): List of visualizations.
    """

    def __init__(self, visualizations):
        self.visualizations = visualizations

    def synchronize(self, time, vehicle_inputs):
        """Synchronize each visualization at the specified time

        Args:
            time (double): the time at which the visualization should synchronize all modules
            vehicle_inputs (WAVehicleInputs): vehicle inputs
        """
        for vis in self.visualizations:
            vis.synchronize(time, vehicle_inputs)

    def advance(self, step):
        """Advance the state of each managed visualization

        Args:
            step (double): the time step at which the visualization should be advanced
        """
        for vis in self.visualizations:
            vis.advance(step)

    def is_ok(self):
        """Verifies the visualization is running properly.

        Returns:
            bool: Whether the visualization is running correctly.
        """
        for vis in self.visualizations:
            if not vis.is_ok():
                return False
        return True


class _MatplotlibVehicle:
    """Plotter class for a matplotlib vehicle"""

    def __init__(self, properties):
        body_Lf = properties["Body Distance to Front"]
        body_Lr = properties["Body Distance to Rear"]
        body_width = properties["Body Width"]
        body_length = properties["Body Length"]
        tire_width = properties["Tire Width"]
        tire_diameter = properties["Tire Diameter"]

        self.Lf = properties["Tire Distance to Front"]
        self.Lr = properties["Tire Distance to Rear"]
        self.track_width = properties["Track Width"]
        self.wheelbase = self.Lf + self.Lr

        # Visualization shapes
        self.outline = np.array(
            [
                [-body_Lr, body_Lf, body_Lf, -body_Lr, -body_Lr],
                [
                    body_width / 2,
                    body_width / 2,
                    -body_width / 2,
                    -body_width / 2,
                    body_width / 2,
                ],
                [1, 1, 1, 1, 1],
            ]
        )

        wheel = np.array(
            [
                [
                    tire_diameter,
                    -tire_diameter,
                    -tire_diameter,
                    tire_diameter,
                    tire_diameter,
                ],
                [-tire_width, -tire_width, tire_width, tire_width, -tire_width],
                [1, 1, 1, 1, 1],
            ]
        )

        self.rr_wheel = np.copy(wheel)
        self.rl_wheel = np.copy(wheel)
        self.rl_wheel[1, :] *= -1

        self.fr_wheel = np.copy(wheel)
        self.fl_wheel = np.copy(wheel)
        self.fl_wheel[1, :] *= -1

    def initialize(self, ax, cabcolor="-k", wheelcolor="-k"):
        """Initialize plotting for the matplotlib vehicle"""
        cab, = ax.plot(
            np.array(self.outline[0, :]).flatten(),
            np.array(self.outline[1, :]).flatten(),
            cabcolor,
        )
        fr, = ax.plot(
            np.array(self.fr_wheel[0, :]).flatten(),
            np.array(self.fr_wheel[1, :]).flatten(),
            wheelcolor,
        )
        rr, = ax.plot(
            np.array(self.rr_wheel[0, :]).flatten(),
            np.array(self.rr_wheel[1, :]).flatten(),
            wheelcolor,
        )
        fl, = ax.plot(
            np.array(self.fl_wheel[0, :]).flatten(),
            np.array(self.fl_wheel[1, :]).flatten(),
            wheelcolor,
        )
        rl, = ax.plot(
            np.array(self.rl_wheel[0, :]).flatten(),
            np.array(self.rl_wheel[1, :]).flatten(),
            wheelcolor,
        )
        self.mat_vehicle = (cab, fr, rr, fl, rl)

    def transform(self, entity, x, y, yaw, alpha=0, x_offset=0, y_offset=0):
        """Helper function to transfrom a numpy entity by the specified values"""
        T = np.array(
            [[np.cos(yaw), np.sin(yaw), x],
             [-np.sin(yaw), np.cos(yaw), y], [0, 0, 1]]
        )
        T = T @ np.array(
            [
                [np.cos(alpha), np.sin(alpha), x_offset],
                [-np.sin(alpha), np.cos(alpha), y_offset],
                [0, 0, 1],
            ]
        )
        return T.dot(entity)

    def update(self, x, y, yaw, steering):
        """Update the state of the matplotlib representation of the vehicle"""

        yaw *= -1
        steering *= -1

        # Update the position of each entity
        fr_wheel = self.transform(
            self.fr_wheel,
            x,
            y,
            yaw,
            alpha=steering,
            x_offset=self.Lf,
            y_offset=-self.track_width,
        )
        fl_wheel = self.transform(
            self.fl_wheel,
            x,
            y,
            yaw,
            alpha=steering,
            x_offset=self.Lf,
            y_offset=self.track_width,
        )
        rr_wheel = self.transform(
            self.rr_wheel, x, y, yaw, x_offset=-self.Lr, y_offset=-self.track_width
        )
        rl_wheel = self.transform(
            self.rl_wheel, x, y, yaw, x_offset=-self.Lr, y_offset=self.track_width
        )
        outline = self.transform(self.outline, x, y, yaw)

        (cab, fr, rr, fl, rl) = self.mat_vehicle

        cab.set_ydata(np.array(outline[1, :]).flatten())
        cab.set_xdata(np.array(outline[0, :]).flatten())
        fr.set_ydata(np.array(fr_wheel[1, :]).flatten())
        fr.set_xdata(np.array(fr_wheel[0, :]).flatten())
        rr.set_ydata(np.array(rr_wheel[1, :]).flatten())
        rr.set_xdata(np.array(rr_wheel[0, :]).flatten())
        fl.set_ydata(np.array(fl_wheel[1, :]).flatten())
        fl.set_xdata(np.array(fl_wheel[0, :]).flatten())
        rl.set_ydata(np.array(rl_wheel[1, :]).flatten())
        rl.set_xdata(np.array(rl_wheel[0, :]).flatten())


class _MatplotlibSimpleDashboard:
    """Simple dashboard with selected information that will be displayed within a matplotlib plot window"""

    def initialize(self, ax):
        """Initialize the annotation"""

        bbox_props = dict(boxstyle="round", fc="w", ec="0.5", alpha=0.9)
        self.annotation = ax.annotate(
            "",
            xy=(0.97, 0.7),
            xytext=(0, 10),
            xycoords=("axes fraction", "figure fraction"),
            textcoords="offset points",
            size=10,
            ha="right",
            va="bottom",
            bbox=bbox_props,
        )

    def update(self, vehicle_inputs, time, speed):
        """Update the annotation"""
        text = (
            f"Time :: {time:.2f}\n"
            f"Steering :: {vehicle_inputs.steering:.2f}\n"
            f"Throttle :: {vehicle_inputs.throttle:.2f}\n"
            f"Braking :: {vehicle_inputs.braking:.2f}\n"
            f"Speed :: {speed:.2f}"
        )
        self.annotation.set_text(text)


class _MatplotlibPlotter(ABC):
    """Base class for matplotlib visualization"""

    class Event:
        def __init__(self, name, value):
            self.name = name
            self.value = value

    def __init__(self, mat_vehicle, dashboard, static=False, padding=25, xlim=(-50, 50), ylim=(50, -50)):
        self.mat_vehicle = mat_vehicle
        self.dashboard = dashboard

        self.vehicle_inputs = WAVehicleInputs()
        self.time = 0

        self.static = static
        self.padding = padding
        self.xlim = xlim
        self.ylim = ylim

        # Asynchronous Matplotlib events
        self.events = dict()

    def _initialize_plot(self):
        """Initialize a matplotlib plot and pass the axes to the vehicle and dashboard"""
        # Initial plotting setup
        self.fig, self.ax = plt.subplots(figsize=(8, 8))

        # Plot styling
        self.ax.set_xlim(*self.xlim)
        self.ax.set_ylim(*self.ylim)

        self.ax.grid(True)
        self.ax.set_aspect("equal", adjustable="box")

        # Initialize the plotters
        self.mat_vehicle.initialize(self.ax)
        self.dashboard.initialize(self.ax)

    def synchronize(self, time, vehicle_inputs):
        """Saves values to be visualized in matplotlib."""
        self.vehicle_inputs = vehicle_inputs
        self.time = time

    @abstractmethod
    def advance(self, step, state):
        pass

    @abstractmethod
    def register_key_press_event(self, callback):
        pass

    @abstractmethod
    def is_alive(self):
        pass

    @abstractmethod
    def plot(self, *args, **kwargs):
        pass

    def _update_axes(self, x, y):
        self.ax.set_xlim(x - self.padding, x + self.padding)
        self.ax.set_ylim(y + self.padding, y - self.padding)


class _MatplotlibSinglePlotter(_MatplotlibPlotter):
    """Single process plotter for sequential visualization. Can be used in jupyter"""

    def __init__(self, mat_vehicle, dashboard, is_jupyter=False, **kwargs):
        super().__init__(mat_vehicle, dashboard, **kwargs)

        self._initialize_plot()

        self.is_jupyter = is_jupyter

    def advance(self, step, state):
        # Extract values
        x, y, yaw, v = state

        # Update the simulation elements
        self.mat_vehicle.update(x, y, yaw, self.vehicle_inputs.steering)
        self.dashboard.update(self.vehicle_inputs, self.time, v)

        if not self.static:
            self._update_axes(x, y)

        # Update the plot
        if self.is_jupyter:
            display(self.fig)
            clear_output(wait=True)
        else:
            plt.pause(1e-9)

    def register_key_press_event(self, callback):
        self.fig.canvas.mpl_connect('key_press_event', self.key_press)
        self.events['key_press_event'] = callback

    def key_press(self, event):
        if event.name in self.events:
            self.events[event.name](event.key)

    def is_alive(self):
        return plt.fignum_exists(self.fig.number)

    def plot(self, *args, **kwargs):
        self.ax.plot(*args, **kwargs)


class _MatplotlibMultiPlotter(_MatplotlibPlotter):
    """Plotter used for situations where multiprocessing is desired. Faster than sequential."""

    class _SimulationState:
        def __init__(self, state, vehicle_inputs, time):
            self.state = state
            self.vehicle_inputs = vehicle_inputs
            self.time = time

        def __call__(self):
            return self.state, self.vehicle_inputs, self.time

    class _PlotCall:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    def __init__(self, mat_vehicle, dashboard, asynchronous=False, **kwargs):
        super().__init__(mat_vehicle, dashboard, **kwargs)

        self.asynchronous = asynchronous

        self.events['close_event'] = self.close

        # Initialize multiprocessing objects
        if plt.get_backend() == "MacOSX":
            set_start_method("spawn", force=True)

        # Communication pipelines between processes
        self.event_sender, self.event_receiver = Pipe()
        self.queue = Queue(1)

        # Barrier that will wait for initialization to occur
        self.barrier = Barrier(2)

        self.p = Process(target=self.run)
        self.p.start()

        # Wait for the matplotlib window to initialize
        self.barrier.wait()

    def advance(self, step, state):
        """Advance the state of the matplotlib window"""
        if self.event_receiver.poll():
            event = self.event_receiver.recv()
            if event.name in self.events:
                self.events[event.name](event.value)

        # Send the state to the other process to visualize
        state = self._SimulationState(state, self.vehicle_inputs, self.time)
        if not self.queue._closed:
            if self.asynchronous and self.queue.empty():
                self.queue.put_nowait(state)
            else:
                self.queue.put(state)

    def run(self):
        """Multiprocess starter method. Will initialize matplotlib and setup FuncAnimation"""
        # Initialize matplotlib
        self._initialize_plot()
        self.fig.canvas.mpl_connect('close_event', self.handle_close)

        # Initialize animation
        anim = FuncAnimation(self.fig, self.update, interval=10)

        # Synchronize with main process
        self.barrier.wait()

        plt.show()

        # Flush the queue
        if not self.queue.empty():
            self.queue.get_nowait()

    def update(self, i):
        """Called at a specific interval by FuncAnimation. Will update matplotlib window"""
        state = self.queue.get()
        if isinstance(state, self._SimulationState):
            (x, y, yaw, v), vehicle_inputs, time = state()

            self.mat_vehicle.update(x, y, yaw, vehicle_inputs.steering)
            self.dashboard.update(vehicle_inputs, time, v)

            if not self.static:
                self._update_axes(x, y)
        elif isinstance(state, str):
            if state == 'key_press_event':
                self.fig.canvas.mpl_connect('key_press_event', self.key_press)
        elif isinstance(state, self._PlotCall):
            self.ax.plot(*state.args, **state.kwargs)

    def handle_close(self, event):
        """Callback from matplotlib when a plot is closed"""
        self.event_sender.send(self.Event(event.name, ''))

    def register_key_press_event(self, callback):
        self.queue.put('key_press_event')
        self.events['key_press_event'] = callback

    def key_press(self, event):
        self.event_sender.send(self.Event(event.name, event.key))

    def is_alive(self):
        return self.p.is_alive()

    def plot(self, *args, **kwargs):
        self.queue.put(self._PlotCall(*args, **kwargs))

    def close(self, event):
        self.queue.close()
        self.event_receiver.close()
        self.event_sender.close()
        self.p.join()

    def __del__(self):
        """Destructor"""
        if hasattr(self, "p") and self.p.is_alive():
            self.p.join()


class WAMatplotlibVisualization(WAVisualization):
    """Matplotlib visualizer of the simulator world and the vehicle

    Args:
        vehicle (WAVehicle): vehicle to render in the matplotlib plot window
        system (WASystem): system used to grab certain parameters of the simulation
        plotter_type (str, optional): Type of plotter. "single" for single threaded, "multi" for multi threaded (fastest), "jupyter" if jupyter is used. Defaults to "single".

    Attributes:
        render_steps (int): steps between which the visualization should update
        vehicle (WAVehicle): vehicle to render in the matplotlib plot window
        system (WASystem): system used to grab certain parameters of the simulation
        mat_vehicle (_MatplotlibVehicle): a wrapper class used to create a visual representation of the vehicle in matplotlib
        dashboard (_MatplotlibSimpleDashboard): a wrapper class used to create a visual dashboard on the matplotlib window
        plotter (_MatplotlibPlotter): a wrapper class that handles matplotlib visualization and possible single/multi threading

    Raises:
        ValueError: plotter_type isn't recognized
    """

    def __init__(self, vehicle, system, plotter_type="single", **kwargs):
        self.render_steps = int(
            np.ceil(system.render_step_size / system.step_size))

        self.vehicle = vehicle
        self.system = system

        self.mat_vehicle = _MatplotlibVehicle(self.vehicle.vis_properties)
        self.dashboard = _MatplotlibSimpleDashboard()

        if plotter_type == "multi":
            self.plotter = _MatplotlibMultiPlotter(
                self.mat_vehicle, self.dashboard, **kwargs)
        elif plotter_type == "single" or plotter_type == "jupyter":
            self.plotter = _MatplotlibSinglePlotter(
                self.mat_vehicle, self.dashboard, is_jupyter=plotter_type == 'jupyter', **kwargs)
        else:
            raise ValueError(
                f"Unknown plotter type: {plotter_type}. Must be multi, single or jupyter")

    def synchronize(self, time, vehicle_inputs):
        """Synchronize the vehicle inputs to the values in this visualization

        Will just set class members

        Args:
            time (double): time at which to update the vehicle to
            vehicle_inputs (WAVehicleInputs): vehicle inputs
        """
        self.plotter.synchronize(time, vehicle_inputs)

    def advance(self, step):
        """Advance the state of the visualization by the specified step

        Will only call update if the scene should be rendered given the render step

        Args:
            step (double): step size to update the visualization by
        """
        if self.system.get_step_number() % self.render_steps == 0:
            self.plotter.advance(step, self.vehicle.get_simple_state())

    def is_ok(self):
        """Checks if the rendering process is still alive

        Returns:
            bool: whether the simulation is still alive
        """
        return self.plotter.is_alive()

    def register_key_press_event(self, callback):
        """Register a key press callback with the matplotlib visualization

        Args:
            callback (function): the callback to invoke when a key is pressed
        """
        self.plotter.register_key_press_event(callback)

    def plot(self, *args, **kwargs):
        """Pass plot info to the underyling plotter"""
        self.plotter.plot(*args, **kwargs)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# Other imports
import numpy as np


class WAVector(np.ndarray):
    """Wrapper of a numpy array to allow for x,y,z properties

    Usage:
        > p1 = WAVector([1, 2])
        > p2 = WAVector([4, 5])
        > p1 + p2
        WAVector([5, 7])
    """
    def __new__(cls, input_array=(0, 0)):
        """Creates a new Point object.

        In python, when an object is created, __init__ and __new__ are called.

        Args:
            cls (class): provided by compiler
            input_array (array_like): Defaults to 2d origin
        """
        obj = np.asarray(input_array).view(cls)
        return obj

    def __eq__(self, other):
        """Check if two vectors are equal

        Args:
            other (wa.WAVector): the vector to cross check with

        Returns:
            bool: whether the values of the vectors are equal
        """
        return np.array_equal(self, other)

    def __ne__(self, other):
        """Check if two vectors are not equal

        Args:
            other (wa.WAVector): the vector to cross check with

        Returns:
            bool: whethe values of the vectors are not equal
        """
        return not np.array_equal(self, other)

    def __iter__(self):
        """Iterate over the x,y,z properties

        Yields:
            float: the x, y or z value
        """
        for x in np.nditer(self):
            yield x.item()

    def __sub__(self, other):
        """Substract a vector from this vector

        Args:
            other (wa.WAVector): the other vector to substract from this vector

        Returns:
            wa.WAVector: the result
        """
        if isinstance(other, int) or isinstance(other, float):
            return np.subtract(self, other)

        n = max(len(self), len(other))
        for i in range(n):
            try:
                self[i]
            except IndexError:
                self = np.append(self, 0)

            try:
                other[i]
            except IndexError:
                other = np.append(other, 0)

        return np.subtract(self, other)

    @property
    def x(self):
        """Get first element

        Raises:
            Exception: exit if undefined

        Returns:
            any: the 1st element
        """
        try:
            return self[0]
        except IndexError:
            raise Exception('Point :: x element does not exist. Exitting...')

    @x.setter
    def x(self, value):
        """Set the first element

        Args:
            value (any): the value to set the first element to

        Raises:
            Exception: Exits if undefined
        """
        try:
            self[0] = value
        except IndexError:
            raise Exception('Point :: x element does not exist. Exitting...')

    @property
    def y(self):
        """Get the second element

        Raises:
            Exception: Exits if undefined

        Returns:
            any: the second element
        """
        try:
            return self[1]
        except IndexError:
            raise Exception('Point :: y element does not exist. Exitting...')

    @y.setter
    def y(self, value):
        """Set the second element

        Args:
            value (any): the value to set the second element to

        Raises:
            Exception: Exits if undefined
        """
        try:
            self[1] = value
        except IndexError:
            raise Exception('Point :: y element does not exist. Exitting...')

    @property
    def z(self):
        """Get the third element

        Raises:
            Exception: Exits if undefined

        Returns:
            any: the third element
        """
        try:
            return self[2]
        except IndexError:
            raise Exception('Point :: z element does not exist. Exitting...')

    @z.setter
    def z(self, value):
        """Set the third element

        Args:
            value (any): the value to set the third element to

        Raises:
            Exception: Exits if undefined
        """
        try:
            self[2] = value
        except IndexError:
            raise Exception('Point :: z element does not exist. Exitting...')

    def length(self):
        """Get the euclidean length of this vector

        Returns:
            float: the euclidean length
        """
        return np.linalg.norm(self)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class


class WASimulation(ABC):
    def __init__(
        self,
        system,
        environment,
        vehicle,
        visualization,
        controller,
        record_filename=None,
    ):
        """A manager for a simulation. Advances and synchronizes simulation modules.

        The simulation object is used primarily for cleaning up main functions. The Run method
        should be used primarily, as the use case for this simulator is mainly for the use of the
        controllers.

        Args:
            system (WASystem): describes the simulation system
            environment (WAEnvironment): describes the simulation environment
            vehicle (WAVehicle): describes the simulation vehicle
            visualization (WAVisualization): describes the simulation visualization
            controller (WAController): describes the simulation controller
            record_filename (str, optional): the filename to store saved simulation data. Defaults to None.

        Attributes:
            system (WASystem): describes the simulation system
            environment (WAEnvironment): describes the simulation environment
            vehicle (WAVehicle): describes the simulation vehicle
            visualization (WAVisualization): describes the simulation visualization
            controller (WAController): describes the simulation controller
            record_filename (str): the filename to store saved simulation data
        """

        self.system = system
        self.environment = environment
        self.vehicle = vehicle
        self.visualization = visualization
        self.controller = controller

        self.record_filename = record_filename
        if self.record_filename:
            # Overwrite current files
            with open(self.record_filename, "w") as f:
                pass

    def record(self, filename):
        """Save simulation state. Only saves the vehicle simple state

        Args:
            filename (str): file to append info to
        """
        with open(filename, "a+") as f:
            x, y, yaw, v = self.vehicle.get_simple_state()
            f.write(f"{x},{y},{yaw},{v}")

    def synchronize(self, time):
        """Synchronize each simulation module.

        Will pass information between simulation modules

        Args:
            time (double): the time to synchronize the simulation to
        """
        vehicle_inputs = self.controller.get_inputs()

        self.controller.synchronize(time)
        self.vehicle.synchronize(time, vehicle_inputs)
        self.environment.synchronize(time)

        if self.visualization:
            # Will synchronize the visualization if necessary
            self.visualization.synchronize(time, vehicle_inputs)

    def advance(self, step):
        """Advance each simulation module.

        Args:
            step (double): the step size to update each module by
        """
        self.system.advance()

        self.environment.advance(step)
        self.vehicle.advance(step)
        self.controller.advance(step)

        if self.visualization:
            # Will advance the visualization if necessary
            self.visualization.advance(step)

    def run(self):
        """Run the simulation

        Can be used to clear up main functions.
        """
        while self.is_ok():
            time = self.system.get_sim_time()

            self.synchronize(time)
            self.advance(self.system.step_size)

            if self.record_filename:
                self.record(self.record_filename)

    def is_ok(self):
        """Verifies the simulation is running as expected.

        Checks the visualization if it's still active. Should probably check other
        modules for any errors.

        Returns:
            bool: Whether the simulation running as expected
        """
        return self.visualization.is_ok() if self.visualization else True



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""


from abc import ABC, abstractmethod  # Abstract Base Class

# Other imports
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import splprep, splev
from scipy.spatial.distance import cdist


def load_waypoints_from_csv(filename, **kwargs):
    """Get data points from a csv file. 

    Should be structured as "x,y,z\nx,y,z...". See NumPy.loadtxt for more info on arguments.

    Args:
        filename (str): file to open and read data from

    Returns:
        np.ndarray: a nxm array with each data point in each row
    """
    return np.loadtxt(filename, **kwargs)


def calc_path_length_cummulative(x, y):
    """Get the cummulative distance along a path provided the given x and y position values

    Args:
        x (np.ndarray): x coordinates
        y (np.ndarray): y coordinates

    Returns:
        np.ndarray: the cummulative distance along the path
    """
    return np.cumsum(np.linalg.norm(np.diff(np.column_stack((x, y)), axis=0), axis=1))


def calc_path_curvature(dx, dy, ddx, ddy):
    """Calculate the curvature of a path at each point

    Args:
        dx (np.ndarray): first x derivative
        dy (np.ndarray): first y derivative
        ddx (np.ndarray): second x derivative
        ddy (np.ndarray): second y derivative

    Returns:
        np.ndarray: the curvature at each point
    """
    return (dx * ddy - dy * ddx) / (dx ** 2 + dy ** 2) ** (3 / 2)


class WAPath:
    """Base Path object. To be used to generate paths or trajectories for path planning and/or path following"""

    @abstractmethod
    def calc_closest_point(self, pos):
        """Calcualte the closest point on the path from the passed position

        Args:
            pos (wa.WAVector): the position to find the closest point on the path to

        Returns:
            wa.WAVector: the closest point on the path
        """
        pass

    @abstractmethod
    def plot(self):
        """Plot the path. Most likely plotter is matplotlib, but technically anything can be used."""
        pass


class WASplinePath(WAPath):
    def __init__(self, waypoints, num_points=100, smoothness=0.0, is_closed=True):
        """Spline path implemented with SciPy's splprep and splev methods

        Args:
            waypoints (np.ndarray): the waypoints to fit the spline to
            num_points (int, optional): number of points to interpolate. Defaults to 100.
            smoothness (float, optional): how fit to each point the spline should be. will hit all points by default. Defaults to 0.0.
            is_closed (bool, optional): Is the path a closed loop. Defaults to True.

        Raises:
            TypeError: the waypoints array type is not as expected
        """
        # Check points type and shape
        if isinstance(waypoints, list):
            waypoints = np.array(waypoints)
        elif not isinstance(waypoints, np.ndarray):
            raise TypeError(
                'waypoints type is not recognized. List or NumPy array required.')

        # Store the waypoints
        self.waypoints = waypoints

        # Interpolate the path
        tck, u = splprep(waypoints.T, s=smoothness, per=is_closed)
        u_new = np.linspace(u.min(), u.max(), num_points)

        # Evaluate the interpolation to get values
        self.x, self.y = splev(u_new, tck, der=0)  # position
        self.dx, self.dy = splev(u_new, tck, der=1)  # first derivative
        self.ddx, self.ddy = splev(u_new, tck, der=2)  # second derivative

        # Variables for tracking path
        self.last_index = None

    def calc_closest_point(self, pos):
        """Calcualte the closest point on the path from the passed position

        Args:
            pos (wa.WAVector): the position to find the closest point on the path to

        Returns:
            wa.WAVector: the closest point on the path
        """
        if len(pos) == 3:
            pos = pos[:2]

        dist = cdist(np.column_stack((self.x, self.y)), [pos])
        idx, = np.argmin(dist, axis=0)
        return self.x[idx], self.y[idx], idx

    def plot(self, *args, show=True, **kwargs):
        """Plot the path

        Args:
            show (bool, optional): show the plot window. Defaults to True.
        """
        plt.plot(self.x, self.y, *args, **kwargs)
        if show:
            plt.show()

    def calc_length_cummulative(self):
        """Get the cummulative distance along the path

        Returns:
            np.ndarray: Cummulative distance along the path
        """
        return calc_path_length_cummulative(self.x, self.y)

    def calc_curvature(self):
        """Get the curvature at each point on the path

        Returns:
            np.ndarray: Curvature at each point on the path
        """
        return calc_path_curvature(self.dx, self.dy, self.ddx, self.ddy)


class WABezierPath(WAPath):

    def __init__(self):
        pass

    def calc_closest_point(self, pos):
        """Calcualte the closest point on the path from the passed position

        Args:
            pos (wa.WAVector): the position to find the closest point on the path to

        Returns:
            wa.WAVector: the closest point on the path
        """
        pass

    def plot(self):
        """Plot the path."""
        pass



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.environment import WAEnvironment
from wa_simulator.chrono.terrain import WAChronoTerrain

# Chrono specific imports
import pychrono as chrono


class WAChronoEnvironment(WAEnvironment):
    """The environment wrapper that's responsible for holding Chrono assets and the terrain

    TODO: Add assets from file

    Args:
        filename (str): the json specification file describing the environment
        system (WASystem): the wa system that wraps a ChSystem
        terrain (ChTerrain, optional): a chrono terrain. Will create one if not passed. Defaults to None.

    Attributes:
        terrain (WAChronoTerrain): the wrapper for the chrono terrain
    """

    # Global filenames for environment models
    EGP_ENV_MODEL_FILE = "environments/ev_grand_prix.json"
    IAC_ENV_MODEL_FILE = "environments/iac.json"

    def __init__(self, filename, system, terrain=None):
        if terrain is None:
            terrain = WAChronoTerrain(filename, system)

        self.terrain = terrain

    def synchronize(self, time):
        """Synchronize the environment. Will just synchronize the terrain

        Args:
            time ([type]): [description]
        """
        self.terrain.synchronize(time)

    def advance(self, step):
        """Advance the environment by the step size. Will just advance the terrain.

        Args:
            step (double): the step size to update the environment by
        """
        self.terrain.advance(step)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import pathlib

# Grab the data folder in the root of this repo
DATA_DIRECTORY = str(pathlib.Path(__file__).absolute().parent / "data")


def get_wa_data_file(filename):
    """Get the absolute path to the file passed

    Args:
            filename (str): file relative to the data folder to get the absolute path for

    Returns:
            str: the absolute path of the file
    """
    return str(pathlib.Path(DATA_DIRECTORY) / filename)


def set_wa_data_directory(path):
    """Set the data path

    Args:
        path (str): relative (or absolute) path where the data is stored
    """
    global DATA_DIRECTORY

    DATA_DIRECTORY = str(pathlib.Path(path).resolve())



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.system import WASystem

# Chrono specific imports
import pychrono as chrono


class WAChronoSystem(WASystem):
    """Chrono system wrapper.

    Performs dynamics automatically

    Args:
        step_size (double): simulation step size
        render_step_size (double, optional): render step size. Defaults to 2e-2.
        contact_method (str, optional): the contact method to use (NSC or SMC). Defaults to "NSC".

    Raises:
        TypeError: verify the contact method is a string
    """

    def __init__(self, step_size, render_step_size=2e-2, contact_method="NSC"):
        super().__init__(step_size, render_step_size)

        if not isinstance(contact_method, str):
            raise TypeError("Contact method must be of type str")

        if contact_method == "NSC":
            system = chrono.ChSystemNSC()
        elif contact_method == "SMC":
            system = chrono.ChSystemSMC()
        system.Set_G_acc(chrono.ChVectorD(0, 0, -9.81))
        system.SetSolverType(chrono.ChSolver.Type_BARZILAIBORWEIN)
        system.SetSolverMaxIterations(150)
        system.SetMaxPenetrationRecoverySpeed(4.0)
        self.system = system

    def advance(self):
        """Advance the simulation.

        Will update the dynamics of the chrono simulation
        """
        self.step_number += 1
        self.system.DoStepDynamics(self.step_size)

    def get_sim_time(self):
        """Get the simulation time.

        Will get the time from the wrapped ChSystem

        Returns:
            double: the simulation time
        """
        self.time = self.system.GetChTime()
        return self.time


"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.controller import WAController, WAVehicleInputs

# Chrono imports
import pychrono.vehicle as veh

# Other imports
import numpy as np


class WAIrrlichtController(WAController):
    """Wrapper for a ChIrrGuiDriver.

    Uses the keyboard in the irrlicht window to control the car. Window must be active
    to work. Currently not supported on Mac when the anaconda build is used

    Args:
        visualization (WAIrrlichtVisualization): the irrlicht visualization
        system (WASystem): the system
    """

    def __init__(self, visualization, system):
        # Create the interactive driver
        driver = veh.ChIrrGuiDriver(visualization.app)

        render_step_size = system.render_step_size

        # Set the time response for steering and throttle keyboard inputs.
        steering_time = 1.0  # time to go from 0 to +1 (or from 0 to -1)
        throttle_time = 1.0  # time to go from 0 to +1
        braking_time = 0.3  # time to go from 0 to +1
        driver.SetSteeringDelta(render_step_size / steering_time)
        driver.SetThrottleDelta(render_step_size / throttle_time)
        driver.SetBrakingDelta(render_step_size / braking_time)

        driver.Initialize()

        self.driver = driver

    def synchronize(self, time):
        """Synchronize the irrlicht driver at the specified time

        Args:
            time (double): time to synchronize at
        """
        self.driver.Synchronize(time)

    def advance(self, step):
        """Advance the irrlicht driver by the specified step

        Args:
            step (double): step to advance by
        """
        self.driver.Advance(step)

    def get_inputs(self):
        """Get the vehicle inputs

        Overrides base class method. Converts a veh.DriverInputs to a WAVehicleInputs

        Returns:
                The input class
        """
        inputs = self.driver.GetInputs()
        return WAVehicleInputs(
            inputs.m_steering,
            inputs.m_throttle,
            inputs.m_braking,
        )


"""
Requires the `PyChrono <https://projectchrono.org/pychrono/>`_ library.
"""

from .._import import _import, _get_files
from .. import *

try:
    # Try to import chrono
    # Will through exception if failed
    import pychrono as chrono
    import pychrono.vehicle as veh
    import pathlib
except Exception as e:
    raise e

# Set the chrono data directory to in-repo data directory
CHRONO_DATA_DIRECTORY = str(pathlib.Path(DATA_DIRECTORY) / "chrono" / " ")[:-1]
CHRONO_VEH_DATA_DIRECTORY = str(
    pathlib.Path(DATA_DIRECTORY) / "chrono" / "vehicle" / " "
)[:-1]

chrono.SetChronoDataPath(CHRONO_DATA_DIRECTORY)
veh.SetDataPath(CHRONO_VEH_DATA_DIRECTORY)


for d in _get_files(__file__):
    _import(d, globals())

del pathlib, _import, _get_files



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.vehicle import WAVehicle

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh


def ReadVehicleModelFile(filename):
    """Read a json specification file to get additional file names to be loaded into ChVehicle classes

    Args:
        filename (str): the json specification file with Vehicle, Powertrain and Tire input models

    Returns:
        tuple: returns each json specification file for the vehicle, powertrain and tire
    """
    import json

    full_filename = veh.GetDataFile(filename)

    with open(full_filename) as f:
        j = json.load(f)

    if "Vehicle" not in j:
        print("'Vehicle' not present in the passed json file.")
        exit()
    vehicle_filename = veh.GetDataFile(j["Vehicle"]["Input File"])

    if "Powertrain" not in j:
        print("'Powertrain' not present in the passed json file.")
        exit()
    powertrain_filename = veh.GetDataFile(j["Powertrain"]["Input File"])

    if "Tire" not in j:
        print("'Tire' not present in the passed json file.")
        exit()
    tire_filename = veh.GetDataFile(j["Tire"]["Input File"])

    return vehicle_filename, powertrain_filename, tire_filename


def CreateTireFromJSON(tire_filename):
    """Creates a tire from a tire file

    Args:
        tire_filename (str): the tire json specification file

    Returns:
        ChTire: the created tire
    """
    import json

    with open(tire_filename) as f:
        j = json.load(f)

    if "Type" not in j:
        print("'Type' not present in the passed json file.")
        exit()

    if j["Type"] != "Tire":
        print("Passed json file is not a tire file.")
        exit()

    if "Template" not in j:
        print("'Template' not present in the passed json file.")
        exit()

    tire_type = j["Template"]
    if tire_type == "TMeasyTire":
        return veh.TMeasyTire(tire_filename)
    elif tire_type == "RigidTire":
        return veh.RigidTire(tire_filename)
    else:
        print(f"'{tire_type}' not a recognized tire type.")
        exit()


class WAChronoVehicle(WAVehicle):
    """Chrono vehicle wrapper

    Args:
        filename (str): json file specification file
        system (WAChronoSystem): the system used to run the simulation
        env (WAEnvironment): the environment with a terrain
        initLoc (chrono.ChVectorD, optional): the inital location of the vehicle. Defaults to chrono.ChVectorD(0, 0, 0.5).
        initRot (chrono.ChQuaternionD, optional): the initial orientation of the vehicle. Defaults to chrono.ChQuaternionD(1, 0, 0, 0).

    Attributes:
        vehicle (ChVehicle): a chrono vehicle that this class essentially wraps
        terrain (ChTerrain): a terrain that the vehicle interacts with
    """

    # Global filenames for vehicle models
    GO_KART_MODEL_FILE = "GoKart/GoKart.json"
    IAC_VEH_MODEL_FILE = "IAC/IAC.json"

    def __init__(
        self,
        filename,
        system,
        env,
        initLoc=chrono.ChVectorD(0, 0, 0.5),
        initRot=chrono.ChQuaternionD(1, 0, 0, 0),
    ):
        super().__init__("vehicles/GoKart/GoKart_KinematicBicycle.json")

        # Get the filenames
        vehicle_file, powertrain_file, tire_file = ReadVehicleModelFile(
            filename)

        # Create the vehicle
        vehicle = veh.WheeledVehicle(system.system, vehicle_file)

        # Initialize the vehicle
        vehicle.Initialize(chrono.ChCoordsysD(initLoc, initRot))

        # Set the visualization components for the vehicle
        vehicle.SetChassisVisualizationType(veh.VisualizationType_PRIMITIVES)
        vehicle.SetSuspensionVisualizationType(
            veh.VisualizationType_PRIMITIVES)
        vehicle.SetSteeringVisualizationType(veh.VisualizationType_PRIMITIVES)
        vehicle.SetWheelVisualizationType(veh.VisualizationType_NONE)

        # Create the powertrain
        # Assumes a SimplePowertrain
        powertrain = veh.SimplePowertrain(powertrain_file)
        vehicle.InitializePowertrain(powertrain)

        # Create and initialize the tires
        for axle in vehicle.GetAxles():
            tireL = CreateTireFromJSON(tire_file)
            vehicle.InitializeTire(
                tireL, axle.m_wheels[0], veh.VisualizationType_PRIMITIVES)
            tireR = CreateTireFromJSON(tire_file)
            vehicle.InitializeTire(
                tireR, axle.m_wheels[1], veh.VisualizationType_PRIMITIVES)

        self.vehicle = vehicle
        self.terrain = env.terrain.terrain

    def advance(self, step):
        """Perform a dynamics update

        Args:
            step (double): time step to update the vehicle by
        """
        self.vehicle.Advance(step)

    def synchronize(self, time, vehicle_inputs):
        """Synchronize the vehicle with the vehicle inputs at the passed time

        Args:
            time (double): time to synchronize the simulation to
            vehicle_inputs (WAVehicleInputs): the vehicle inputs
        """
        if not isinstance(self.terrain, veh.ChTerrain):
            raise TypeError("Synchronize: Terrain has not been set.")

        d = veh.Inputs()
        d.m_steering = vehicle_inputs.steering
        d.m_throttle = vehicle_inputs.throttle
        d.m_braking = vehicle_inputs.braking
        vehicle_inputs = d

        self.vehicle.Synchronize(time, vehicle_inputs, self.terrain)

    def get_simple_state(self):
        """Get a simple state representation of the vehicle.

        Returns:
            tuple: (x position, y position, yaw about the Z, speed)
        """
        pos = self.vehicle.GetVehiclePos()
        return (
            pos.x,
            pos.y,
            self.vehicle.GetVehicleRot().Q_to_Euler123().z,
            self.vehicle.GetVehicleSpeed(),
        )
